<?php
/**
 * Bootstrap
 *
 * PHP version 5
 *
 * Copyright (C) Villanova University 2010.
 * Copyright (C) project swissbib, University Library Basel, Switzerland
 * Copyright (C) Leipzig University Library 2015.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  finc
 * @author   Guenter Hipler  <guenter.hipler@unibas.ch>
 * @author   Frank Morgner   <morgner.f@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     http://www.swissbib.org
 */
use Zend\Mvc\Service\ServiceManagerConfig;
use Zend\ServiceManager\ServiceManager;
use Zend\Stdlib\ArrayUtils;

/**
 * Bootstrap
 *
 * @category VuFind
 * @package  finc
 * @author   Guenter Hipler  <guenter.hipler@unibas.ch>
 * @author   Frank Morgner   <morgner.f@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     http://vufind.org
 */
class Bootstrap
{
    /**
     * ServiceManager
     *
     * @var ServiceManager
     */
    protected static $serviceManager;

    /**
     * Config
     *
     * @var static config
     */
    protected static $config;

    /**
     * Bootstrap
     *
     * @var static bootstrap
     */
    protected static $bootstrap;

    /**
     * Init
     *
     * @return void
     */
    public static function init()
    {
        // Load the user-defined test configuration file, if it exists;
        // otherwise, load
        if (is_readable(FINC_TESTS_PATH . '/config.php')) {
            $testConfig = include FINC_TESTS_PATH . '/config.php';
        } else {
            $testConfig = include FINC_TESTS_PATH . '/config.php.dist';
        }

        $zf2ModulePaths = [];
        if (isset($testConfig['module_listener_options']['module_paths'])) {
            $modulePaths = $testConfig['module_listener_options']['module_paths'];
            foreach ($modulePaths as $modulePath) {
                if (($path = static::findParentPath($modulePath))) {
                    $zf2ModulePaths[] = $path;
                }
            }
        }
        $zf2ModulePaths = implode(PATH_SEPARATOR, $zf2ModulePaths) . PATH_SEPARATOR;
        $zf2ModulePaths .= getenv('ZF2_MODULES_TEST_PATHS') ?: (defined('ZF2_MODULES_TEST_PATHS') ? ZF2_MODULES_TEST_PATHS : '');
        static::initAutoloader();
        // use ModuleManager to load this module and it's dependencies
        $baseConfig = [
            'module_listener_options' => [
                'module_paths' => explode(PATH_SEPARATOR, $zf2ModulePaths),
            ],
        ];
        self::initEnvironment();
        $config = ArrayUtils::merge($baseConfig, $testConfig);
        $serviceManager = new ServiceManager(new ServiceManagerConfig());
        $serviceManager->setService('ApplicationConfig', $config);
        $serviceManager->get('ModuleManager')->loadModules();
        static::$serviceManager = $serviceManager;
        static::$config         = $config;
    }

    /**
     * InitVuFind
     *
     * @return void
     */
    public function initEnvironment()
    {
        define('APPLICATION_ENV', 'development');
        define('FINC_TEST_FIXTURES', realpath(FINC_TESTS_PATH . '/fixtures'));
    }
    /**
     * GetServiceManager
     *
     * @return ServiceManager
     */
    public static function getServiceManager()
    {
        return static::$serviceManager;
    }
    /**
     * GetConfig
     *
     * @return self::config
     * @static
     */
    public static function getConfig()
    {
        return static::$config;
    }
    /**
     * InitAutoloader
     *
     * @return void
     */
    protected static function initAutoloader()
    {
        $vendorPath = static::findParentPath('vendor');
        if (is_readable($vendorPath . '/autoload.php')) {
            include $vendorPath . '/autoload.php';
            $loader = new \Composer\Autoload\ClassLoader();
            $loader->add('VuFindTest', APPLICATION_PATH . '/module/VuFind/tests/unit-tests/src');
            $loader->add('VuFindTest', APPLICATION_PATH . '/module/VuFind/src');
            $loader->add('fincTest', FINC_TESTS_PATH . '/unit-tests/src');
            $loader->add('VuFindTest', APPLICATION_PATH . '/module/VuFind/src/VuFindTest');
            $loader->register();
        } else {
            throw new RuntimeException('Unable initialize autoloading.');
        }
    }
    /**
     * FindParentPath
     *
     * @param String $path Path
     *
     * @return bool|string
     */
    protected static function findParentPath($path)
    {
        $dir         = FINC_TESTS_PATH;
        $previousDir = '.';
        while (!is_dir($dir . '/' . $path)) {
            $dir = dirname($dir);
            if ($previousDir === $dir) {
                return false;
            }
            $previousDir = $dir;
        }
        return $dir . '/' . $path;
    }
}

// Set flag that we're in test mode
define('VUFIND_PHPUNIT_RUNNING', 1);

// Set path to this module
define('VUFIND_PHPUNIT_MODULE_PATH', __DIR__);

// Define default search backend identifier
defined('DEFAULT_SEARCH_BACKEND')
|| define('DEFAULT_SEARCH_BACKEND', 'Solr');

// Define path to local override directory
defined('LOCAL_OVERRIDE_DIR')
|| define(
    'LOCAL_OVERRIDE_DIR',
    (getenv('VUFIND_LOCAL_DIR') ? getenv('VUFIND_LOCAL_DIR') : '')
);

// Define path to cache directory
defined('LOCAL_CACHE_DIR')
|| define(
    'LOCAL_CACHE_DIR',
    (getenv('VUFIND_CACHE_DIR')
        ? getenv('VUFIND_CACHE_DIR')
        : (
        strlen(LOCAL_OVERRIDE_DIR) > 0 ? LOCAL_OVERRIDE_DIR . '/cache' : '')
    )
);

define('APPLICATION_PATH', realpath(dirname(__DIR__) . '/../..'));
define('FINC_TESTS_PATH', __DIR__);
chdir(APPLICATION_PATH);

Bootstrap::init();